use redis::{AsyncCommands, JsonAsyncCommands};
use sqlx::{query, query_as, types::Json};

use super::models::Account;
use crate::{ctx::Ctx, error::NekrochanError, perms::PermissionWrapper};

impl Account {
    pub async fn create(
        ctx: &Ctx,
        username: String,
        password: String,
    ) -> Result<Self, NekrochanError> {
        let account =
            query_as("INSERT INTO accounts (username, password) VALUES ($1, $2) RETURNING *")
                .bind(&username)
                .bind(password)
                .fetch_one(ctx.db())
                .await?;

        ctx.cache()
            .json_set(format!("accounts:{username}"), ".", &account)
            .await?;

        Ok(account)
    }

    pub async fn read(ctx: &Ctx, username: String) -> Result<Option<Self>, NekrochanError> {
        let account: Option<String> = ctx
            .cache()
            .json_get(format!("accounts:{username}"), ".")
            .await?;

        let account = match account {
            Some(json) => Some(serde_json::from_str(&json)?),
            None => None,
        };

        Ok(account)
    }

    pub async fn read_all(ctx: &Ctx) -> Result<Vec<Self>, NekrochanError> {
        let accounts = query_as("SELECT * FROM accounts ORDER BY owner DESC, created DESC")
            .fetch_all(ctx.db())
            .await?;

        Ok(accounts)
    }

    pub async fn update_password(&self, ctx: &Ctx, password: String) -> Result<(), NekrochanError> {
        query("UPDATE accounts SET password = $1 WHERE username = $2")
            .bind(&password)
            .bind(&self.username)
            .execute(ctx.db())
            .await?;

        ctx.cache()
            .json_set(format!("accounts:{}", self.username), "password", &password)
            .await?;

        Ok(())
    }

    pub async fn update_permissions(
        &self,
        ctx: &Ctx,
        permissions: u64,
    ) -> Result<(), NekrochanError> {
        query("UPDATE accounts SET permissions = $1 WHERE username = $2")
            .bind(Json(permissions))
            .bind(&self.username)
            .execute(ctx.db())
            .await?;

        ctx.cache()
            .json_set(
                format!("accounts:{}", self.username),
                "permissions",
                &permissions,
            )
            .await?;

        Ok(())
    }

    pub async fn update_owner(&self, ctx: &Ctx, owner: bool) -> Result<(), NekrochanError> {
        query("UPDATE accounts SET owner = $1 WHERE username = $2")
            .bind(owner)
            .bind(&self.username)
            .execute(ctx.db())
            .await?;

        ctx.cache()
            .json_set(format!("accounts:{}", self.username), "owner", &owner)
            .await?;

        Ok(())
    }

    pub async fn delete(&self, ctx: &Ctx) -> Result<(), NekrochanError> {
        query("DELETE FROM accounts WHERE username = $1")
            .bind(&self.username)
            .execute(ctx.db())
            .await?;

        ctx.cache()
            .del(format!("accounts:{}", self.username))
            .await?;

        Ok(())
    }
}

impl Account {
    pub fn perms(&self) -> PermissionWrapper {
        PermissionWrapper::new(self.permissions.0, self.owner)
    }
}
