use anyhow::Error;
use redis::{cmd, AsyncCommands, JsonAsyncCommands};
use sha256::digest;
use sqlx::query_as;

use super::models::{Account, Banner, Board, Post};
use crate::ctx::Ctx;

pub async fn init_cache(ctx: &Ctx) -> Result<(), Error> {
    cmd("FLUSHDB").query_async(&mut ctx.cache()).await?;

    let accounts: Vec<Account> = query_as("SELECT * FROM accounts")
        .fetch_all(ctx.db())
        .await?;

    for account in &accounts {
        ctx.cache()
            .json_set(format!("accounts:{}", account.username), ".", &account)
            .await?;
    }

    let boards: Vec<Board> = query_as("SELECT * FROM boards").fetch_all(ctx.db()).await?;

    for board in &boards {
        ctx.cache()
            .json_set(format!("boards:{}", board.id), ".", board)
            .await?;

        ctx.cache().lpush("board_ids", &board.id).await?;
    }

    let banners: Vec<Banner> = query_as("SELECT * FROM banners")
        .fetch_all(ctx.db())
        .await?;

    for banner in &banners {
        ctx.cache()
            .zadd("banners", serde_json::to_string(banner)?, banner.id)
            .await?;
    }

    cmd("SORT")
        .arg("board_ids")
        .arg("ALPHA")
        .arg("STORE")
        .arg("board_ids")
        .query_async(&mut ctx.cache())
        .await?;

    ctx.cache().set("total_threads", 0).await?;

    for board in &boards {
        let (thread_count,): (i64,) = query_as(&format!(
            "SELECT COUNT(id) FROM posts_{} WHERE thread IS NULL",
            board.id
        ))
        .fetch_one(ctx.db())
        .await?;

        ctx.cache().incr("total_threads", thread_count).await?;
        ctx.cache()
            .set(format!("board_threads:{}", board.id), thread_count)
            .await?;
    }

    for board in &boards {
        let posts = Post::read_all(ctx, board.id.clone()).await?;

        for post in posts {
            let ip_key = format!("by_ip:{}", post.ip);
            let content_key = format!("by_content:{}", digest(post.content_nomarkup));

            let member = format!("{}/{}", post.board, post.id);
            let score = post.created.timestamp_micros();

            ctx.cache().zadd(ip_key, &member, score).await?;
            ctx.cache().zadd(content_key, &member, score).await?;
        }
    }

    Ok(())
}
