use actix_web::{http::StatusCode, ResponseError};
use log::error;
use thiserror::Error;

#[derive(Error, Debug)]
pub enum NekrochanError {
    #[error("Účet '{}' neexistuje.", .0)]
    AccountNotFound(String),
    #[error("Tento ban už byl odvolán.")]
    AlreadyAppealedError,
    #[error("Odvolání můsí mít 1-1000 znaků.")]
    BanAppealFormatError,
    #[error("Žádný takový ban pro tuto IP adresu neexistuje.")]
    BanNotFound,
    #[error("Důvod banu musí mít 1-200 znaků.")]
    BanReasonFormatError,
    #[error("Nástěnka /{}/ je uzamčená.", .0)]
    BoardLockError(String),
    #[error("Jméno nástěnky musí mít 1-32 znaků.")]
    BoardNameFormatError,
    #[error("Nástěnka /{}/ neexistuje.", .0)]
    BoardNotFound(String),
    #[error("Capcode nesmí mít více než 32 znaků.")]
    CapcodeFormatError,
    #[error("Obsah nesmí mít více než 10000 znaků.")]
    ContentFormatError,
    #[error("Popis musí mít 1-128 znaků.")]
    DescriptionFormatError,
    #[error("E-mail nesmí mít více než 256 znaků.")]
    EmailFormatError,
    #[error("Příspěvek musí mít obsah nebo soubor.")]
    EmptyPostError,
    #[error("Chyba při zpracovávání souboru '{}': {}", .0, .1)]
    FileError(String, &'static str),
    #[error("Maximální počet souborů na této nástěnce je {}.", .0)]
    FileLimitError(usize),
    #[error("Tvůj příspěvek vypadá jako spam.")]
    FloodError,
    #[error("Reverzní proxy nevrátilo vyžadovanou hlavičku '{}'.", .0)]
    HeaderError(&'static str),
    #[error("Domovní stránka vznikne po vytvoření nástěnky.")]
    HomePageError,
    #[error("ID musí mít 1-16 znaků.")]
    IdFormatError,
    #[error("Nesprávné řešení CAPTCHA.")]
    IncorrectCaptchaError,
    #[error("Nesprávné přihlašovací údaje.")]
    IncorrectCredentialError,
    #[error("Nesprávné heslo pro příspěvek #{}.", .0)]
    IncorrectPasswordError(i64),
    #[error("Nedostatečná oprávnění.")]
    InsufficientPermissionError,
    #[error("Server se připojil k 41 procentům.")]
    InternalError,
    #[error("Neplatný autentizační token. Vymaž soubory cookie.")]
    InvalidAuthError,
    #[error("Neplatná strana.")]
    InvalidPageError,
    #[error("Obsah musí mít 1-20000 znaků.")]
    NewsContentFormatError,
    #[error("Titulek musí mít 1-100 znaků.")]
    NewsTitleFormatError,
    #[error("Příspěvek musí mít obsah.")]
    NoContentError,
    #[error("Příspěvek musí mít soubor.")]
    NoFileError,
    #[error("Nebyly vybrány žádné příspěvky.")]
    NoPostsError,
    #[error("Pro přístup se musíš přihlásit.")]
    NotLoggedInError,
    #[error("Nadnástěnka nebyla inicializována.")]
    OverboardError,
    #[error("Účet vlastníka nemůže být vymazán.")]
    OwnerDeletionError,
    #[error("Heslo musí mít alespoň 8 znaků.")]
    PasswordFormatError,
    #[error("Jméno nesmí mít více než 32 znaků.")]
    PostNameFormatError,
    #[error("Příspěvek /{}/{} neexistuje.", .0, .1)]
    PostNotFound(String, i64),
    #[error("Vlákno dosáhlo limitu odpovědí.")]
    ReplyLimitError,
    #[error("Hlášení můsí mít 1-200 znaků.")]
    ReportFormatError,
    #[error("Nelze vytvořit odpověď na odpověď.")]
    ReplyReplyError,
    #[error("Na této nástěnce se musí vyplnit CAPTCHA.")]
    RequiredCaptchaError,
    #[error("Tato CAPTCHA neexistuje nebo už byla vyřešena.")]
    SolvedCaptchaError,
    #[error("Toto vlákno je uzamčené.")]
    ThreadLockError,
    #[error("Tento ban nelze odvolat.")]
    UnappealableError,
    #[error("Uživatelské jméno musí mít 1-32 znaků.")]
    UsernameFormatError,
}

impl From<askama::Error> for NekrochanError {
    fn from(e: askama::Error) -> Self {
        error!("Internal server error: {e:#?}");

        Self::InternalError
    }
}

impl From<ipnetwork::IpNetworkError> for NekrochanError {
    fn from(e: ipnetwork::IpNetworkError) -> Self {
        error!("Internal server error: {e:#?}");

        Self::InternalError
    }
}

impl From<jsonwebtoken::errors::Error> for NekrochanError {
    fn from(e: jsonwebtoken::errors::Error) -> Self {
        error!("Internal server error: {e:#?}");

        Self::InternalError
    }
}

impl From<pwhash::error::Error> for NekrochanError {
    fn from(e: pwhash::error::Error) -> Self {
        error!("Internal server error: {e:#?}");

        Self::InternalError
    }
}

impl From<regex::Error> for NekrochanError {
    fn from(e: regex::Error) -> Self {
        error!("Internal server error: {e:#?}");

        Self::InternalError
    }
}

impl From<redis::RedisError> for NekrochanError {
    fn from(e: redis::RedisError) -> Self {
        error!("Internal server error: {e:#?}");

        Self::InternalError
    }
}

impl From<serde_json::Error> for NekrochanError {
    fn from(e: serde_json::Error) -> Self {
        error!("Internal server error: {e:#?}");

        Self::InternalError
    }
}

impl From<serde_qs::Error> for NekrochanError {
    fn from(e: serde_qs::Error) -> Self {
        error!("Internal server error: {e:#?}");

        Self::InternalError
    }
}

impl From<sqlx::Error> for NekrochanError {
    fn from(e: sqlx::Error) -> Self {
        let overboard_err = match e.as_database_error() {
            Some(e) => e.message() == "relation \"overboard\" does not exist",
            None => false,
        };

        if overboard_err {
            Self::OverboardError
        } else {
            error!("{e:#?}");

            Self::InternalError
        }
    }
}

impl From<std::io::Error> for NekrochanError {
    fn from(e: std::io::Error) -> Self {
        error!("Internal server error: {e:#?}");

        Self::InternalError
    }
}

impl From<std::net::AddrParseError> for NekrochanError {
    fn from(e: std::net::AddrParseError) -> Self {
        error!("Internal server error: {e:#?}");

        Self::InternalError
    }
}

impl<T> From<std::sync::PoisonError<T>> for NekrochanError {
    fn from(_: std::sync::PoisonError<T>) -> Self {
        error!("CAPTCHA RwLock got poisoned or something.");

        Self::InternalError
    }
}

impl From<tokio::task::JoinError> for NekrochanError {
    fn from(e: tokio::task::JoinError) -> Self {
        error!("Internal server error: {e:#?}");

        Self::InternalError
    }
}

impl ResponseError for NekrochanError {
    fn status_code(&self) -> StatusCode {
        match self {
            NekrochanError::AccountNotFound(_) => StatusCode::NOT_FOUND,
            NekrochanError::AlreadyAppealedError => StatusCode::BAD_REQUEST,
            NekrochanError::BanAppealFormatError => StatusCode::BAD_REQUEST,
            NekrochanError::BanNotFound => StatusCode::NOT_FOUND,
            NekrochanError::BanReasonFormatError => StatusCode::BAD_REQUEST,
            NekrochanError::BoardLockError(_) => StatusCode::FORBIDDEN,
            NekrochanError::BoardNameFormatError => StatusCode::BAD_REQUEST,
            NekrochanError::BoardNotFound(_) => StatusCode::NOT_FOUND,
            NekrochanError::CapcodeFormatError => StatusCode::BAD_REQUEST,
            NekrochanError::ContentFormatError => StatusCode::BAD_REQUEST,
            NekrochanError::DescriptionFormatError => StatusCode::BAD_REQUEST,
            NekrochanError::EmailFormatError => StatusCode::BAD_REQUEST,
            NekrochanError::EmptyPostError => StatusCode::BAD_REQUEST,
            NekrochanError::FileError(_, _) => StatusCode::UNPROCESSABLE_ENTITY,
            NekrochanError::FileLimitError(_) => StatusCode::BAD_REQUEST,
            NekrochanError::FloodError => StatusCode::TOO_MANY_REQUESTS,
            NekrochanError::HeaderError(_) => StatusCode::BAD_GATEWAY,
            NekrochanError::HomePageError => StatusCode::NOT_FOUND,
            NekrochanError::IdFormatError => StatusCode::BAD_REQUEST,
            NekrochanError::IncorrectCaptchaError => StatusCode::UNAUTHORIZED,
            NekrochanError::IncorrectCredentialError => StatusCode::UNAUTHORIZED,
            NekrochanError::IncorrectPasswordError(_) => StatusCode::UNAUTHORIZED,
            NekrochanError::InsufficientPermissionError => StatusCode::FORBIDDEN,
            NekrochanError::InternalError => StatusCode::INTERNAL_SERVER_ERROR,
            NekrochanError::InvalidAuthError => StatusCode::UNAUTHORIZED,
            NekrochanError::InvalidPageError => StatusCode::BAD_REQUEST,
            NekrochanError::NewsContentFormatError => StatusCode::BAD_REQUEST,
            NekrochanError::NewsTitleFormatError => StatusCode::BAD_REQUEST,
            NekrochanError::NoContentError => StatusCode::BAD_REQUEST,
            NekrochanError::NoFileError => StatusCode::BAD_REQUEST,
            NekrochanError::NoPostsError => StatusCode::BAD_REQUEST,
            NekrochanError::NotLoggedInError => StatusCode::UNAUTHORIZED,
            NekrochanError::OverboardError => StatusCode::INTERNAL_SERVER_ERROR,
            NekrochanError::OwnerDeletionError => StatusCode::FORBIDDEN,
            NekrochanError::PasswordFormatError => StatusCode::BAD_REQUEST,
            NekrochanError::PostNameFormatError => StatusCode::BAD_REQUEST,
            NekrochanError::PostNotFound(_, _) => StatusCode::NOT_FOUND,
            NekrochanError::ReplyLimitError => StatusCode::FORBIDDEN,
            NekrochanError::ReplyReplyError => StatusCode::BAD_REQUEST,
            NekrochanError::ReportFormatError => StatusCode::BAD_REQUEST,
            NekrochanError::RequiredCaptchaError => StatusCode::UNAUTHORIZED,
            NekrochanError::SolvedCaptchaError => StatusCode::BAD_REQUEST,
            NekrochanError::ThreadLockError => StatusCode::FORBIDDEN,
            NekrochanError::UnappealableError => StatusCode::BAD_REQUEST,
            NekrochanError::UsernameFormatError => StatusCode::BAD_REQUEST,
        }
    }
}
