use actix_web::{post, web::Data, HttpRequest, HttpResponse};
use serde::Deserialize;
use std::fmt::Write;

use crate::{
    ctx::Ctx,
    db::models::{Ban, Board},
    error::NekrochanError,
    qsform::QsForm,
    web::{
        actions::{get_posts_from_ids, ActionTemplate},
        ban_response,
        tcx::{ip_from_req, TemplateCtx},
        template_response,
    },
};

#[derive(Deserialize)]
pub struct ReportPostsForm {
    #[serde(default)]
    pub posts: Vec<String>,
    pub report_reason: String,
}

#[post("/actions/report-posts")]
pub async fn report_posts(
    ctx: Data<Ctx>,
    req: HttpRequest,
    QsForm(form): QsForm<ReportPostsForm>,
) -> Result<HttpResponse, NekrochanError> {
    let tcx = TemplateCtx::new(&ctx, &req).await?;
    let (reporter_ip, reporter_country) = ip_from_req(&req)?;
    let bans = Ban::read_by_ip(&ctx, reporter_ip).await?;

    if let Some(ban) = bans.get(&None) {
        if !(tcx.perms.owner() || tcx.perms.bypass_bans()) {
            return ban_response(&ctx, &req, ban.clone()).await;
        }
    }

    let boards = Board::read_all_map(&ctx).await?;
    let posts = get_posts_from_ids(&ctx, form.posts).await;

    let mut response = String::new();
    let mut posts_reported = 0;

    let reason = form.report_reason.trim();

    if reason.is_empty() || reason.len() > 200 {
        return Err(NekrochanError::ReportFormatError);
    }

    for post in &posts {
        let board = &boards[&post.board];

        if bans.contains_key(&Some(board.id.clone()))
            && !(tcx.perms.owner() || tcx.perms.bypass_bans())
        {
            writeln!(&mut response, "[Chyba] Jsi zabanován z /{}/.", board.id).ok();
            continue;
        }

        if board.config.0.locked && !(tcx.perms.owner() || tcx.perms.bypass_board_lock()) {
            writeln!(
                &mut response,
                "[Chyba] {}",
                NekrochanError::BoardLockError(board.id.clone())
            )
            .ok();

            continue;
        }

        if post
            .reports
            .iter()
            .any(|report| report.reporter_ip == reporter_ip)
        {
            writeln!(
                &mut response,
                "[Chyba] Příspěvek #{} jsi už nahlásil.",
                post.id
            )
            .ok();
            continue;
        }

        post.create_report(
            &ctx,
            reason.to_owned(),
            reporter_country.clone(),
            reporter_ip,
        )
        .await?;

        posts_reported += 1;
    }

    if posts_reported != 0 {
        writeln!(
            &mut response,
            "[Úspěch] Nahlášeny příspěvky: {posts_reported}"
        )
        .ok();
    }

    let template = ActionTemplate { tcx, response };

    template_response(&template)
}
