use actix_web::{post, web::Data, HttpRequest, HttpResponse};
use chrono::{Duration, Utc};
use ipnetwork::IpNetwork;
use serde::Deserialize;
use std::{collections::HashSet, fmt::Write};

use crate::{
    ctx::Ctx,
    db::models::Ban,
    error::NekrochanError,
    qsform::QsForm,
    web::{
        actions::{get_posts_from_ids, ActionTemplate},
        tcx::{account_from_auth, TemplateCtx},
        template_response,
    },
};

#[derive(Deserialize)]
pub struct JannyPostActionsForm {
    #[serde(default)]
    pub posts: Vec<String>,
    #[serde(rename = "staff_remove_posts")]
    pub remove_posts: Option<String>,
    #[serde(rename = "staff_remove_files")]
    pub remove_files: Option<String>,
    #[serde(rename = "staff_toggle_spoiler")]
    pub toggle_spoiler: Option<String>,
    pub toggle_sticky: Option<String>,
    pub toggle_lock: Option<String>,
    pub ban_user: Option<String>,
    pub global_ban: Option<String>,
    pub unappealable_ban: Option<String>,
    pub ban_reason: Option<String>,
    pub ban_duration: Option<i64>,
    pub ban_range: Option<String>,
}

#[post("/actions/staff-post-actions")]
pub async fn staff_post_actions(
    ctx: Data<Ctx>,
    req: HttpRequest,
    QsForm(form): QsForm<JannyPostActionsForm>,
) -> Result<HttpResponse, NekrochanError> {
    let tcx = TemplateCtx::new(&ctx, &req).await?;
    let account = account_from_auth(&ctx, &req).await?;
    let posts = get_posts_from_ids(&ctx, form.posts).await;

    let mut response = String::new();

    let mut posts_removed = 0;
    let mut files_removed = 0;
    let mut spoilers_toggled = 0;
    let mut stickies_toggled = 0;
    let mut locks_toggled = 0;
    let mut bans_issued = 0;

    for post in &posts {
        if (form.remove_posts.is_some()
            || form.remove_files.is_some()
            || form.toggle_spoiler.is_some()
            || form.toggle_sticky.is_some()
            || form.toggle_lock.is_some())
            && !(account.perms().owner() || account.perms().manage_posts())
        {
            writeln!(
                &mut response,
                "[Chyba] Nemáš oprávnění spravovat příspěvky."
            )
            .ok();

            continue;
        }

        if form.remove_posts.is_some() {
            post.delete(&ctx).await?;
            posts_removed += 1;
        }

        if form.remove_files.is_some() {
            post.delete_files(&ctx).await?;
            files_removed += post.files.0.len();
        }

        if form.toggle_spoiler.is_some() {
            post.update_spoiler(&ctx).await?;
            spoilers_toggled += post.files.0.len();
        }

        if form.toggle_sticky.is_some() {
            post.update_sticky(&ctx).await?;
            stickies_toggled += post.files.0.len();
        }

        if form.toggle_lock.is_some() {
            post.update_lock(&ctx).await?;
            locks_toggled += post.files.0.len();
        }
    }

    let mut already_banned = HashSet::new();

    for post in posts {
        if let (Some(_), Some(ban_reason), Some(ban_duration), Some(ban_range)) = (
            form.ban_user.clone(),
            form.ban_reason.clone(),
            form.ban_duration,
            form.ban_range.clone(),
        ) {
            if !(account.perms().owner() || account.perms().bans()) {
                writeln!(&mut response, "[Chyba] Nemáš oprávnění vydat ban.").ok();
                continue;
            }

            if already_banned.contains(&post.ip) {
                continue;
            }

            let account = account.username.clone();

            let board = if form.global_ban.is_none() {
                Some(post.board.clone())
            } else {
                None
            };

            let prefix = if post.ip.is_ipv4() {
                match ban_range.as_str() {
                    "lan" => 24,
                    "isp" => 16,
                    _ => 32,
                }
            } else {
                match ban_range.as_str() {
                    "lan" => 48,
                    "isp" => 24,
                    _ => 128,
                }
            };

            let ip_range = IpNetwork::new(post.ip, prefix)?;
            let reason: String = ban_reason.trim().into();

            if reason.is_empty() || reason.len() > 200 {
                return Err(NekrochanError::BanReasonFormatError);
            }

            let appealable = form.unappealable_ban.is_none();

            let expires = if ban_duration == 0 {
                None
            } else {
                Some(Utc::now() + Duration::days(ban_duration))
            };

            Ban::create(&ctx, account, board, ip_range, reason, appealable, expires).await?;

            let content_nomarkup = format!(
                "{}\n\n==(UŽIVATEL BYL ZA TENTO PŘÍSPĚVEK ZABANOVÁN)==",
                post.content_nomarkup
            );

            let content = format!(
                "{}\n\n<span class=\"redtext\">(UŽIVATEL BYL ZA TENTO PŘÍSPĚVEK ZABANOVÁN)</span>",
                post.content
            );

            post.update_content(&ctx, content, content_nomarkup).await?;

            already_banned.insert(post.ip);
            bans_issued += 1;
        }
    }

    if posts_removed != 0 {
        writeln!(
            &mut response,
            "[Úspěch] Odstraněny příspěvky: {posts_removed}"
        )
        .ok();
    }

    if files_removed != 0 {
        writeln!(
            &mut response,
            "[Úspěch] Odstraněny soubory: {files_removed}"
        )
        .ok();
    }

    if spoilers_toggled != 0 {
        writeln!(
            &mut response,
            "[Úspěch] Přepnuty spoilery: {spoilers_toggled}"
        )
        .ok();
    }

    if stickies_toggled != 0 {
        writeln!(
            &mut response,
            "[Úspěch] Připnuta/odepnuta vlákna: {stickies_toggled}"
        )
        .ok();
    }

    if locks_toggled != 0 {
        writeln!(
            &mut response,
            "[Úspěch] Zamčena/odemčena vlákna: {locks_toggled}"
        )
        .ok();
    }

    if bans_issued != 0 {
        writeln!(&mut response, "[Úspěch] Uděleny bany: {bans_issued}").ok();
    }

    let template = ActionTemplate { tcx, response };

    template_response(&template)
}
