use actix_web::{
    get,
    web::{Data, Path, Query},
    HttpRequest, HttpResponse,
};
use askama::Template;
use redis::AsyncCommands;
use serde::Deserialize;

use crate::{
    check_page,
    ctx::Ctx,
    db::models::{Board, Post},
    error::NekrochanError,
    filters, paginate,
    web::{tcx::TemplateCtx, template_response},
};

#[derive(Deserialize)]
pub struct BoardQuery {
    page: i64,
}

#[derive(Template)]
#[template(path = "board.html")]
struct BoardTemplate {
    tcx: TemplateCtx,
    board: Board,
    threads: Vec<(Post, Vec<Post>)>,
    page: i64,
    pages: i64,
}

#[get("/boards/{board}")]
pub async fn board(
    ctx: Data<Ctx>,
    req: HttpRequest,
    path: Path<String>,
    query: Option<Query<BoardQuery>>,
) -> Result<HttpResponse, NekrochanError> {
    let tcx = TemplateCtx::new(&ctx, &req).await?;

    let board = path.into_inner();
    let board = Board::read(&ctx, board.clone())
        .await?
        .ok_or(NekrochanError::BoardNotFound(board))?;

    let count = ctx
        .cache()
        .get(format!("board_threads:{}", board.id))
        .await?;
    let page = query.map_or(1, |q| q.page);
    let pages = paginate(board.config.0.page_size, count);

    check_page(page, pages, board.config.0.page_count)?;

    let mut threads = Vec::new();

    for thread in Post::read_board_page(&ctx, &board, page).await? {
        let replies = thread.read_replies(&ctx).await?;

        threads.push((thread, replies));
    }

    let template = BoardTemplate {
        tcx,
        board,
        threads,
        page,
        pages,
    };

    template_response(&template)
}
