use actix_web::{
    get,
    web::{Data, Query},
    HttpRequest, HttpResponse,
};
use askama::Template;
use redis::AsyncCommands;
use serde::Deserialize;
use std::collections::HashMap;

use crate::{
    check_page,
    ctx::Ctx,
    db::models::{Board, Post},
    error::NekrochanError,
    filters, paginate,
    web::{tcx::TemplateCtx, template_response},
};

#[derive(Deserialize)]
pub struct OverboardQuery {
    page: i64,
}

#[derive(Template)]
#[template(path = "overboard.html")]
struct OverboardTemplate {
    tcx: TemplateCtx,
    boards: HashMap<String, Board>,
    threads: Vec<(Post, Vec<Post>)>,
    page: i64,
    pages: i64,
}

#[get("/overboard")]
pub async fn overboard(
    ctx: Data<Ctx>,
    req: HttpRequest,
    query: Option<Query<OverboardQuery>>,
) -> Result<HttpResponse, NekrochanError> {
    let boards = Board::read_all_map(&ctx).await?;
    let tcx = TemplateCtx::new(&ctx, &req).await?;

    let count = ctx.cache().get("total_threads").await?;
    let page = query.map_or(1, |q| q.page);
    let pages = paginate(15, count);

    check_page(page, pages, None)?;

    let mut threads = Vec::new();

    for thread in Post::read_overboard_page(&ctx, page).await? {
        let replies = thread.read_replies(&ctx).await?;

        threads.push((thread, replies));
    }

    let template = OverboardTemplate {
        tcx,
        boards,
        threads,
        page,
        pages,
    };

    template_response(&template)
}
