use actix_web::{post, web::Data, HttpRequest, HttpResponse};
use serde::Deserialize;

use crate::{
    cfg::BoardCfg, ctx::Ctx, db::models::Board, error::NekrochanError, qsform::QsForm,
    web::tcx::account_from_auth,
};

#[derive(Deserialize)]
pub struct UpdateBoardConfigForm {
    board: String,
    anon_name: String,
    page_size: i64,
    page_count: i64,
    file_limit: usize,
    bump_limit: i32,
    reply_limit: i32,
    locked: Option<String>,
    user_ids: Option<String>,
    flags: Option<String>,
    thread_captcha: String,
    reply_captcha: String,
    require_thread_content: Option<String>,
    require_thread_file: Option<String>,
    require_reply_content: Option<String>,
    require_reply_file: Option<String>,
    antispam: Option<String>,
    antispam_ip: i64,
    antispam_content: i64,
    antispam_both: i64,
}

#[post("/staff/actions/update-board-config")]
pub async fn update_board_config(
    ctx: Data<Ctx>,
    req: HttpRequest,
    QsForm(form): QsForm<UpdateBoardConfigForm>,
) -> Result<HttpResponse, NekrochanError> {
    let account = account_from_auth(&ctx, &req).await?;

    if !(account.perms().owner() || account.perms().board_config()) {
        return Err(NekrochanError::InsufficientPermissionError);
    }

    let board = form.board;
    let board = Board::read(&ctx, board.clone())
        .await?
        .ok_or(NekrochanError::BoardNotFound(board))?;

    let anon_name = form.anon_name;
    let page_size = form.page_size;
    let page_count = form.page_count;
    let file_limit = form.file_limit;
    let bump_limit = form.bump_limit;
    let reply_limit = form.reply_limit;
    let locked = form.locked.is_some();
    let user_ids = form.user_ids.is_some();
    let flags = form.flags.is_some();
    let thread_captcha = form.thread_captcha;
    let reply_captcha = form.reply_captcha;
    let require_thread_content = form.require_thread_content.is_some();
    let require_thread_file = form.require_thread_file.is_some();
    let require_reply_content = form.require_reply_content.is_some();
    let require_reply_file = form.require_reply_file.is_some();
    let antispam = form.antispam.is_some();
    let antispam_ip = form.antispam_ip;
    let antispam_content = form.antispam_content;
    let antispam_both = form.antispam_both;

    let config = BoardCfg {
        anon_name,
        page_size,
        page_count,
        file_limit,
        bump_limit,
        reply_limit,
        locked,
        user_ids,
        flags,
        thread_captcha,
        reply_captcha,
        require_thread_content,
        require_thread_file,
        require_reply_content,
        require_reply_file,
        antispam,
        antispam_ip,
        antispam_content,
        antispam_both,
    };

    board.update_config(&ctx, config).await?;

    let res = HttpResponse::SeeOther()
        .append_header(("Location", "/staff/boards"))
        .finish();

    Ok(res)
}
